import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Optional

import requests

from fmi_weather_client.errors import ServiceError

_LOGGER = logging.getLogger(__name__)

STORED_QUERY_FORECAST = 'fmi::forecast::hirlam::surface::point::multipointcoverage'


def request_observations_by_coordinates(lat: float, lon: float) -> str:
    """
    Get the latest weather information by coordinates.

    :param lat: Latitude (e.g. 25.67087)
    :param lon: Longitude (e.g. 62.39758)
    :return: Latest weather information
    """
    end_time = datetime.utcnow().replace(tzinfo=timezone.utc)
    start_time = end_time - timedelta(minutes=10)
    params = _create_params(start_time, end_time, 10, lat=lat, lon=lon)
    return _send_request(params)


def request_observations_by_place(place: str) -> str:
    """
    Get the latest weather information by place name.

    :param place: Place name (e.g. Kaisaniemi, Helsinki)
    :return: Latest weather information
    """
    end_time = datetime.utcnow().replace(tzinfo=timezone.utc)
    start_time = end_time - timedelta(minutes=10)
    params = _create_params(start_time, end_time, 10, place=place)
    return _send_request(params)


def request_forecast_by_coordinates(lat: float, lon: float, timestep_hours: int = 24) -> str:
    """
    Get the latest forecast by place coordinates

    :param lat: Latitude (e.g. 25.67087)
    :param lon: Longitude (e.g. 62.39758)
    :param timestep_hours: Forecast steps in hours
    :return: Forecast response
    """
    start_time = datetime.utcnow().replace(tzinfo=timezone.utc)
    end_time = start_time + timedelta(days=4)
    timestep = timestep_hours * 60
    params = _create_params(start_time, end_time, timestep, lat=lat, lon=lon)
    return _send_request(params)


def request_forecast_by_place(place: str, timestep_hours: int = 24) -> str:
    """
    Get the latest forecast by place coordinates

    :param place: Place name (e.g. Kaisaniemi,Helsinki)
    :param timestep_hours: Forecast steps in hours
    :return: Forecast response
    """
    start_time = datetime.utcnow().replace(tzinfo=timezone.utc)
    end_time = start_time + timedelta(days=4)
    timestep = timestep_hours * 60
    params = _create_params(start_time, end_time, timestep, place=place)
    return _send_request(params)


def _create_params(start_time: datetime,
                   end_time: datetime,
                   timestep_minutes: int,
                   place: Optional[str] = None,
                   lat: Optional[float] = None,
                   lon: Optional[float] = None) -> Dict[str, Any]:
    """
    Create query parameters
    :param start_time: Start datetime
    :param end_time: End datetime
    :param timestep_minutes: Timestamp minutes
    :param place: Place name
    :param lat: Latitude
    :param lon: Longitude
    :return: Parameters
    """

    if place is None and lat is None and lon is None:
        raise Exception("Missing location parameter")

    params = {
        'service': 'WFS',
        'version': '2.0.0',
        'request': 'getFeature',
        'storedquery_id': STORED_QUERY_FORECAST,
        'timestep': timestep_minutes,
        'starttime': start_time.isoformat(timespec='seconds'),
        'endtime': end_time.isoformat(timespec='seconds')
    }

    if lat is not None and lon is not None:
        params['latlon'] = '%s,%s' % (lat, lon)

    if place is not None:
        params['place'] = place.strip().replace(' ', '')

    return params


def _send_request(params: Dict[str, Any]) -> str:
    """
    Send a request to FMI service and return the body
    :param params: Query parameters
    :return: Response body
    """
    url = 'http://opendata.fmi.fi/wfs'

    _LOGGER.debug("Sending GET to %s with parameters: %s", url, params)
    response = requests.get(url, params=params)

    if response.status_code >= 500:
        raise ServiceError("Invalid FMI service response", {'status_code': response.status_code, 'body': response.text})

    _LOGGER.debug("Received a response from FMI in %s ms", response.elapsed.microseconds / 1000)
    return response.text



from typing import Any, Dict, Optional

import math


def float_or_none(v: Any) -> Optional[float]:
    """
    Get value as float. None if conversion fails.
    :param v: Any value
    :return: Value as float if successful; None otherwise
    """
    try:
        f = float(v)
        if not math.isnan(f):
            return f
        return None
    except (ValueError, TypeError):
        return None


def is_non_empty_forecast(forecast: Dict[str, float]) -> bool:
    for _, value in forecast.items():
        if not math.isnan(value):
            return True

    return False



from fmi_weather_client import http
from fmi_weather_client.models import Weather
from fmi_weather_client.parsers import forecast as forecast_parser


def weather_by_coordinates(lat: float, lon: float) -> Weather:
    """
    Get the latest weather information by coordinates.

    :param lat: Latitude (e.g. 25.67087)
    :param lon: Longitude (e.g. 62.39758)
    :return: Latest weather information
    """
    response = http.request_observations_by_coordinates(lat, lon)
    forecast = forecast_parser.parse_forecast(response)
    weather_state = forecast.forecasts[-1]
    return Weather(forecast.place, forecast.lat, forecast.lon, weather_state)


def weather_by_place_name(name: str) -> Weather:
    """
    Get the latest weather information by place name.

    :param name: Place name (e.g. Kaisaniemi, Helsinki)
    :return: Latest weather information
    """
    response = http.request_observations_by_place(name)
    forecast = forecast_parser.parse_forecast(response)
    weather_state = forecast.forecasts[-1]
    return Weather(forecast.place, forecast.lat, forecast.lon, weather_state)


def forecast_by_place_name(name: str, timestep_hours: int = 24):
    """
    Get the latest forecast by place name
    :param name: Place name
    :param timestep_hours: Hours between forecasts
    :return: Latest forecast
    """
    response = http.request_forecast_by_place(name, timestep_hours)
    return forecast_parser.parse_forecast(response)


def forecast_by_coordinates(lat: float, lon: float, timestep_hours: int = 24):
    """
    Get the latest forecast by coordinates
    :param lat: Latitude (e.g. 25.67087)
    :param lon: Longitude (e.g. 62.39758)
    :param timestep_hours: Hours between forecasts
    :return: Latest forecast
    """
    response = http.request_forecast_by_coordinates(lat, lon, timestep_hours)
    return forecast_parser.parse_forecast(response)



from datetime import datetime, timezone
from typing import Any, Dict, List

import xmltodict

from fmi_weather_client import errors, utils
from fmi_weather_client.errors import NoDataAvailableError
from fmi_weather_client.models import FMIForecast, FMIForecastTime, FMIStation, Forecast, WeatherData


def parse_forecast(body: str):
    """
    Parse FMI forecast response body to dictionary and check errors
    :param body: Forecast response body
    :return: Response body as dictionary
    """
    data = _body_to_dict(body)
    stations = _get_stations(data)
    times = _get_forecast_times(data)
    types = _get_forecast_variable_types(data)
    value_sets = _get_forecast_variable_values(data)

    # Combine values with types
    typed_value_sets: List[Dict[str, float]] = []
    for value_set in value_sets:
        typed_value_set = {}
        for idx, value in enumerate(value_set):
            typed_value_set[types[idx]] = value
        typed_value_sets.append(typed_value_set)

    # Combine typed values with times
    forecasts: List[FMIForecast] = []
    for idx, time in enumerate(times):
        forecasts.append(FMIForecast(time.lat, time.lon, time.timestamp, typed_value_sets[idx]))

    # Build forecast objects
    station_forecasts = []
    for station in stations:
        station_forecast = Forecast(station.name, station.lat, station.lon, [])
        for forecast in forecasts:
            if forecast.lat == station.lat and forecast.lon == station.lon and utils.is_non_empty_forecast(
                    forecast.values):
                station_forecast.forecasts.append(WeatherData(forecast.timestamp, forecast.values))
        station_forecasts.append(station_forecast)

    # I always saw just one station so I guess this is fine
    return station_forecasts[0]


def _body_to_dict(body: str) -> Dict[str, Any]:
    """
    Convert FMI response body to dictionary and check errors
    :param body: FMI response body
    :return: Response as dictionary
    """
    data = xmltodict.parse(body)

    # Check exception response
    if 'ExceptionReport' in data.keys():
        if 'No locations found for the place' in data['ExceptionReport']['Exception']['ExceptionText'][0]:
            raise errors.NoDataAvailableError
        elif 'No data available for' in data['ExceptionReport']['Exception']['ExceptionText'][0]:
            raise errors.NoDataAvailableError
        raise errors.ServiceError(data['ExceptionReport']['Exception']['ExceptionText'][0])

    if 'wfs:member' not in data['wfs:FeatureCollection'].keys():
        raise errors.NoDataAvailableError

    return data


def _get_stations(data: Dict[str, Any]) -> List[FMIStation]:
    """
    Try to get station data from the response. When call is made with place name, there is only one
    station available. For coordinate search there might be more.
    :param data: Response data from FMI as xmltodict object
    :return: List of stations
    """

    def parse_station_data(station_data: Dict[str, Any]) -> FMIStation:
        name = station_data['gml:Point']['gml:name']
        position = station_data['gml:Point']['gml:pos'].split(' ', 1)
        return FMIStation(name,
                          float(position[1]),
                          float(position[0]))

    stations = []
    stations_dict = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']
                         ['om:featureOfInterest']['sams:SF_SpatialSamplingFeature']['sams:shape']
                         ['gml:MultiPoint']['gml:pointMembers'])

    # xmltodict can return a list or an object depending on how many child the element has
    if isinstance(stations_dict, list):
        for station_dict in stations_dict:
            station = parse_station_data(station_dict)
            stations.append(station)
    else:
        stations.append(parse_station_data(stations_dict))

    return stations


def _get_forecast_times(data: Dict[str, Any]) -> List[FMIForecastTime]:
    result = []
    forecast_times_data_list = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']
                                    ['om:result']['gmlcov:MultiPointCoverage']['gml:domainSet']
                                    ['gmlcov:SimpleMultiPoint']['gmlcov:positions'].split('\n'))
    for forecast_time in forecast_times_data_list:
        parts = forecast_time.strip().replace('  ', ' ').split(' ')
        lon = float(parts[0])
        lat = float(parts[1])
        timestamp = datetime.utcfromtimestamp(int(parts[2])).replace(tzinfo=timezone.utc)
        result.append(FMIForecastTime(lat, lon, timestamp))

    return result


def _get_forecast_variable_types(data) -> List[str]:
    result = []
    variable_type_list = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']
                              ['om:result']['gmlcov:MultiPointCoverage']['gmlcov:rangeType']['swe:DataRecord']
                              ['swe:field'])

    for variable_type in variable_type_list:
        result.append(variable_type['@name'])

    return result


def _get_forecast_variable_values(data: Dict[str, Any]) -> List[List[float]]:
    result = []
    forecast_value_sets = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']
                           ['om:result']['gmlcov:MultiPointCoverage']['gml:rangeSet']['gml:DataBlock']
                           ['gml:doubleOrNilReasonTupleList'].split('\n'))

    contains_values = False

    for forecast_value_set in forecast_value_sets:
        forecast_values = forecast_value_set.strip().split(' ')
        value_set = []
        for value in forecast_values:
            value_set.append(float(value))
            if not contains_values and utils.float_or_none(value) is not None:
                contains_values = True

        result.append(value_set)

    if not contains_values:
        raise NoDataAvailableError

    return result



from datetime import datetime
from typing import Optional, List, Dict


class FMIStation:
    def __init__(self, name: str, lat: float, lon: float):
        self.name: str = name
        self.lat: float = lat
        self.lon: float = lon

    def __str__(self):
        return "%s (%s, %s)" % (self.name, self.lat, self.lon)


class FMIForecastTime:
    def __init__(self, lat: float, lon: float, timestamp: datetime):
        self.lat: float = lat
        self.lon: float = lon
        self.timestamp: datetime = timestamp

    def __str__(self):
        return "%s, %s, %s" % (self.lat, self.lon, self.timestamp)


class FMIForecast:
    def __init__(self, lat: float, lon: float, timestamp: datetime, values: Dict[str, float]):
        self.lat: float = lat
        self.lon: float = lon
        self.timestamp: datetime = timestamp
        self.values: Dict[str, float] = values


class Value:
    def __init__(self, value: Optional[float], unit: str):
        self.value: Optional[float] = value
        self.unit: str = unit

    def __str__(self):
        return "%s %s" % (self.value, self.unit)


class WeatherData:
    def __init__(self, timestamp: datetime, values: Dict[str, float]):

        def to_value(o: Dict[str, float], variable_name: str, unit: str) -> Value:
            value = o.get(variable_name, None)
            return Value(value, unit)

        self.time: datetime = timestamp
        self.geopotential_height: Value = to_value(values, 'GeopHeight', 'm')
        self.temperature: Value = to_value(values, 'Temperature', '°C')
        self.pressure: Value = to_value(values, 'Pressure', 'hPa')
        self.humidity: Value = to_value(values, 'Humidity', '%')
        self.wind_direction: Value = to_value(values, 'WindDirection', '°')
        self.wind_speed: Value = to_value(values, 'WindSpeedMS', 'm/s')
        self.wind_u_component: Value = to_value(values, 'WindUMS', 'm/s')
        self.wind_v_component: Value = to_value(values, 'WindVMS', 'm/s')
        self.wind_max: Value = to_value(values, 'MaximumWind', 'm/s')
        self.wind_gust: Value = to_value(values, 'WindGust', 'm/s')
        self.dew_point: Value = to_value(values, 'DewPoint', '°')
        self.cloud_cover: Value = to_value(values, 'TotalCloudCover', '%')
        self.symbol: Value = to_value(values, 'WeatherSymbol3', '')
        self.cloud_low_cover: Value = to_value(values, 'LowCloudCover', '%')
        self.cloud_mid_cover: Value = to_value(values, 'MediumCloudCover', '%')
        self.cloud_high_cover: Value = to_value(values, 'HighCloudCover', '%')
        self.precipitation_amount_1h: Value = to_value(values, 'Precipitation1h', 'mm/h')
        self.precipitation_amount: Value = to_value(values, 'PrecipitationAmount', 'mm')
        self.radiation_global_short_wave_acc: Value = to_value(values, 'RadiationGlobalAccumulation', 'J/m²')
        self.radiation_global_long_wave_acc: Value = to_value(values, 'RadiationLWAccumulation', 'J/m²')
        self.radiation_surface_long_wave_acc: Value = to_value(values, 'RadiationNetSurfaceLWAccumulation', 'J/m²')
        self.radiation_diff_surface_short_wave_acc: Value = to_value(values, 'RadiationDiffuseAccumulation', 'J/m²')
        self.land_sea_mask: Value = to_value(values, 'LandSeaMask', '')


class Weather:
    def __init__(self, place: str, lat: float, lon: float, weather_data: WeatherData):
        self.place: str = place
        self.lat: float = lat
        self.lon: float = lon
        self.data: WeatherData = weather_data


class Forecast:
    def __init__(self, place: str, lat: float, lon: float, forecasts: List[WeatherData]):
        self.place: str = place
        self.lat: float = lat
        self.lon: float = lon
        self.forecasts: List[WeatherData] = forecasts



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fmi-weather-client",
    version="0.0.7",
    author="Mika Hiltunen",
    author_email="saaste@gmail.com",
    description="Library for fetching weather information from Finnish Meteorological Institute (FMI)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/saaste/fmi-weather-client",
    packages=setuptools.find_packages(),
    install_requires=[
        'requests>=2.22.0',
        'xmltodict>=0.12.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6.9',
)



import unittest
from unittest import mock

import fmi_weather_client
import fmi_weather_client.test.test_data as test_data
from fmi_weather_client.errors import NoDataAvailableError, ServiceError


class FMIWeatherTest(unittest.TestCase):

    @mock.patch('requests.get', side_effect=test_data.mock_place_forecast_response)
    def test_get_observation_by_place(self, mock_get):
        observation = fmi_weather_client.weather_by_place_name('Iisalmi')
        self.assertEqual(observation.place, 'Iisalmi')
        self.assertEqual(observation.data.time.timestamp(), 1582027200.0)
        self.assertEqual(observation.data.temperature.value, 2.34)
        self.assertEqual(observation.data.wind_speed.value, 3.52)

    @mock.patch('requests.get', side_effect=test_data.mock_coordinate_forecast_response)
    def test_get_observation_by_coordinates(self, mock_get):
        observation = fmi_weather_client.weather_by_coordinates(63.14343, 27.31317)
        self.assertEqual(observation.place, 'Sauoiva')
        self.assertEqual(observation.data.time.timestamp(), 1582027200.0)
        self.assertEqual(observation.data.temperature.value, 0.37)
        self.assertEqual(observation.data.wind_speed.value, 1.35)

    @mock.patch('requests.get', side_effect=test_data.mock_empty_response)
    def test_empty_response(self, mock_get):
        with self.assertRaises(NoDataAvailableError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_nan_response)
    def test_nan_response(self, mock_get):
        with self.assertRaises(NoDataAvailableError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_no_location_exception_response)
    def test_no_location_exception_response(self, mock_get):
        with self.assertRaises(NoDataAvailableError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_other_exception_response)
    def test_other_exception_response(self, mock_get):
        with self.assertRaises(ServiceError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_place_forecast_response)
    def test_get_forecast_by_place_name(self, mock_get):
        forecast = fmi_weather_client.forecast_by_place_name('Iisalmi')
        self.assertEqual(forecast.place, 'Iisalmi')
        self.assertEqual(forecast.lat, 27.19067)
        self.assertEqual(forecast.lon, 63.55915)
        self.assertEqual(len(forecast.forecasts), 5)
        self.assertEqual(forecast.forecasts[0].temperature.value, 1.21)
        self.assertEqual(forecast.forecasts[1].pressure.value, 974.67)
        self.assertEqual(forecast.forecasts[4].humidity.value, 96.31)

    @mock.patch('requests.get', side_effect=test_data.mock_coordinate_forecast_response)
    def test_get_forecast_by_coordinates(self, mock_get):
        forecast = fmi_weather_client.forecast_by_coordinates(29.742731, 67.583988)
        self.assertEqual(forecast.place, 'Sauoiva')
        self.assertEqual(forecast.lat, 29.74273)
        self.assertEqual(forecast.lon, 67.58399)
        self.assertEqual(len(forecast.forecasts), 5)
        self.assertEqual(forecast.forecasts[0].temperature.value, -2.26)
        self.assertEqual(forecast.forecasts[1].pressure.value, 968.58)
        self.assertEqual(forecast.forecasts[4].humidity.value, 97.0)


if __name__ == '__main__':
    unittest.main()
